/*-
 * %%----------------------------------------------------------------------------------------------
 * Solidify Framework - Solidify Frontend - abstract-app.component.ts
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2023 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

import {
  AfterViewInit,
  ChangeDetectorRef,
  Directive,
  ElementRef,
  HostListener,
  Injector,
  OnInit,
  Renderer2,
  Type,
  ViewChild,
} from "@angular/core";
import {
  Router,
  RouterStateSnapshot,
} from "@angular/router";
import {TranslateService} from "@ngx-translate/core";
import {Navigate} from "@ngxs/router-plugin";
import {Store} from "@ngxs/store";
import {
  combineLatest,
  fromEvent,
  Observable,
} from "rxjs";
import {
  distinctUntilChanged,
  filter,
  map,
  take,
  tap,
} from "rxjs/operators";
import {AppConfigService} from "../../../config/app-config.service";
import {SOLIDIFY_CONSTANTS} from "../../../constants";
import {CookieConsentService} from "../../../cookie-consent/services/cookie-consent.service";
import {LanguagePartialEnum} from "../../../enums/partial/language-partial.enum";
import {LocalStoragePartialEnum} from "../../../enums/partial/local-storage-partial.enum";
import {ThemePartialEnum} from "../../../enums/partial/theme-partial.enum";
import {LocalStorageHelper} from "../../../helpers/local-storage.helper";
import {SessionStorageHelper} from "../../../helpers/session-storage.helper";
import {FrontendVersion} from "../../../models/frontend-version.model";
import {StateModel} from "../../../models/stores/state.model";
import {BreakpointService} from "../../../services/breakpoint.service";
import {GoogleAnalyticsService} from "../../../services/google-analytics.service";
import {LoggingService} from "../../../services/logging.service";
import {MetaService} from "../../../services/meta.service";
import {NotificationService} from "../../../services/notification.service";
import {ScrollService} from "../../../services/scroll.service";
import {SolidifyAppStateModel} from "../../../stores/abstract/app/app-state.model";
import {SolidifyAppAction} from "../../../stores/abstract/app/app.action";
import {SolidifyAppState} from "../../../stores/abstract/app/app.state";
import {
  isFalse,
  isInstanceOf,
  isNotNullNorUndefined,
  isNullOrUndefined,
  isString,
  isTrue,
  isTruthyObject,
} from "../../../tools/is/is.tool";
import {ExtendEnum} from "../../../types/extend-enum.type";
import {SolidifyObject} from "../../../types/solidify-object.type";
import {DateUtil} from "../../../utils/date.util";
import {SsrUtil} from "../../../utils/ssr.util";
import {MemoizedUtil} from "../../../utils/stores/memoized.util";
import {AbstractInternalContainer} from "../abstract-internal/abstract-internal.container";

declare let gtag: (name: string, id: string, option: SolidifyObject) => void;

// @dynamic
@Directive()
export abstract class AbstractAppComponent extends AbstractInternalContainer implements OnInit, AfterViewInit {

  protected readonly _ATTRIBUTE_SRC_DIFFER_KEY: string = "src-differ";
  protected readonly _ATTRIBUTE_SRC: string = "src";

  private static readonly _THEME_ATTRIBUTE_NAME: string = "theme";
  private static readonly _DARK_MODE_ATTRIBUTE_NAME: string = "dark-mode";
  private static readonly _UI_ATTRIBUTE_NAME: string = "ui";
  logo: string;

  urlStateObs: Observable<RouterStateSnapshot> = this._store.select((state: StateModel) => state.router.state);
  appLanguageObs: Observable<ExtendEnum<LanguagePartialEnum>>;
  isLoggedObs: Observable<boolean>;
  isApplicationInitializedObs: Observable<boolean>;
  isServerOfflineObs: Observable<boolean>;
  isUnableToLoadAppObs: Observable<boolean>;
  themeObs: Observable<ExtendEnum<ThemePartialEnum>>;
  darkModeObs: Observable<boolean>;
  isTouchInterfaceObs: Observable<boolean>;

  frontendVersionObs: Observable<FrontendVersion>;

  ignoreGrid: boolean = false;
  displayCart: boolean = false;
  isMaintenanceModeActive: boolean = false;

  @ViewChild("main")
  mainElementRef: ElementRef;

  @HostListener("window:beforeunload", ["$event"])
  leaveWithoutSaveNotification($event: Event): void {
    const preventExit = MemoizedUtil.selectSnapshot(this._store, this.environment.appState, state => state.preventExit);
    const ignorePreventExit = MemoizedUtil.selectSnapshot(this._store, this.environment.appState, state => state.ignorePreventLeavePopup);
    if (isFalse(ignorePreventExit) && isTrue(preventExit)) {
      $event.returnValue = true;
    }
  }

  get sessionStorageHelper(): typeof SessionStorageHelper {
    return SessionStorageHelper;
  }

  get localStorageHelper(): typeof LocalStorageHelper {
    return LocalStorageHelper;
  }

  constructor(protected readonly _appState: Type<SolidifyAppState<SolidifyAppStateModel>>,
              protected readonly _store: Store,
              protected readonly _router: Router,
              protected readonly _translate: TranslateService,
              protected readonly _renderer: Renderer2,
              protected readonly _injector: Injector,
              protected readonly _notificationService: NotificationService,
              protected readonly _loggingService: LoggingService,
              public readonly breakpointService: BreakpointService,
              protected readonly _scrollService: ScrollService,
              protected readonly _googleAnalyticsService: GoogleAnalyticsService,
              protected readonly _cookieConsentService: CookieConsentService,
              protected readonly _appConfigService: AppConfigService,
              protected readonly _changeDetector: ChangeDetectorRef,
              protected readonly _metaService: MetaService,
  ) {
    super(_injector);
    new DateUtil(); // Allow to initiate moment in DateUtil
    this.environment.appState = this._appState;
    this.frontendVersionObs = MemoizedUtil.select(this._store, this.environment.appState, state => state.frontendVersion);
    this.appLanguageObs = MemoizedUtil.select(this._store, this.environment.appState, state => state.appLanguage);
    this.isLoggedObs = MemoizedUtil.select(this._store, this.environment.appState, state => state.isLoggedIn);
    this.isApplicationInitializedObs = MemoizedUtil.select(this._store, this.environment.appState, state => state.isApplicationInitialized);
    this.isServerOfflineObs = MemoizedUtil.select(this._store, this.environment.appState, state => state.isServerOffline);
    this.isUnableToLoadAppObs = MemoizedUtil.select(this._store, this.environment.appState, state => state.isUnableToLoadApp);
    this.themeObs = MemoizedUtil.select(this._store, this.environment.appState, state => state.theme);
    this.darkModeObs = MemoizedUtil.select(this._store, this.environment.appState, state => state.darkMode);
    this.isTouchInterfaceObs = MemoizedUtil.select(this._store, this.environment.appState, state => state.isTouchInterface);
    this._store.dispatch(new SolidifyAppAction.InitApplication());
    this._setOverrideCssStyleSheet();
    if (SsrUtil.isBrowser) {
      this.subscribe(this._observeThemeChange());
      this.subscribe(this._observeDarkModeChange());
    } else {
      this._setTheme();
      this._setFavicon();
      this._setImageToolbar();
    }
    this.themeChange(this.environment.theme);
    this._initDarkMode();
    this.subscribe(this._listenLanguageChange());
    this.subscribe(this._listenAppReady());
    if (SsrUtil.isBrowser) {
      this._cookieConsentService.openBannerIfNeverAnswer();
    }
    this._googleAnalyticsService.init(this._renderer);
  }

  // Override if necessary
  protected _isLoggedInRoot(): boolean {
    return false;
  }

  ngOnInit(): void {
    super.ngOnInit();

    if (SsrUtil.isBrowser) {
      this.subscribe(this._observeToUpdateMaintenanceMode());
      this.subscribe(this.isTouchInterfaceObs, isTouchInterface => this._setTouchInterface(isTouchInterface));
      this.subscribe(this._observeTouchEvent());
    }
    this.subscribe(this._observeCurrentUrl());
  }

  private _observeCurrentUrl(): Observable<string> {
    this._findMetaToUpdate();
    return this.urlStateObs
      .pipe(
        filter(urlState => isNotNullNorUndefined(urlState)),
        map(urlState => urlState.url),
        distinctUntilChanged(),
        tap(urlRequested => {
          this._findMetaToUpdate(urlRequested);
        }),
      );
  }

  get urlRequestedSnapshot(): string | undefined {
    return this._store.selectSnapshot((state: StateModel) => state.router.state)?.url;
  }

  get isApplicationInitializedSnapshot(): boolean {
    return MemoizedUtil.selectSnapshot(this._store, this.environment.appState, state => state.isApplicationInitialized);
  }

  private _findMetaToUpdate(urlRequested: string = this.urlRequestedSnapshot): void {
    if (!this.isApplicationInitializedSnapshot) {
      return;
    }
    this._metaService.findMetaToUpdate(urlRequested);
  }

  private _observeToUpdateMaintenanceMode(): Observable<any> {
    this.isMaintenanceModeActive = isTrue(this.environment.maintenanceMode);
    return combineLatest([this.isApplicationInitializedObs, this._appConfigService.environmentUpdatedObs])
      .pipe(
        map(([isInitialized, environment]) => {
          this.isMaintenanceModeActive = isTrue(this.environment.maintenanceMode);
          this._changeDetector.detectChanges();
          return this.isMaintenanceModeActive;
        }),
      );
  }

  ngAfterViewInit(): void {
    super.ngAfterViewInit();
    if (isTruthyObject(this.mainElementRef)) {
      this._scrollService.init(this.mainElementRef);
    }
  }

  private _observeTouchEvent(): Observable<Event> {
    return fromEvent(SsrUtil.window?.document, "touchstart")
      .pipe(
        take(1),
        tap(event => this._store.dispatch(new SolidifyAppAction.SetIsTouchInterface(true)),
        ),
      );
  }

  useLanguage(language: ExtendEnum<LanguagePartialEnum>): void {
    this._store.dispatch(new SolidifyAppAction.ChangeAppLanguage(language));
  }

  navigate(route: string | Navigate): void {
    if (isString(route)) {
      this._store.dispatch(new Navigate([route]));
    } else if (isInstanceOf(route, Navigate)) {
      this._store.dispatch(route);
    }
  }

  private _setTheme(): void {
    this._renderer.setAttribute(SsrUtil.window?.document.body, AbstractAppComponent._THEME_ATTRIBUTE_NAME, String(this.environment.theme));
  }

  private _setDarkMode(enabled: boolean): void {
    this._renderer.setAttribute(SsrUtil.window?.document.body, AbstractAppComponent._DARK_MODE_ATTRIBUTE_NAME, enabled ? "true" : "false");
  }

  private _setTouchInterface(isTouchInterface: boolean): void {
    this._renderer.setAttribute(SsrUtil.window?.document.body, AbstractAppComponent._UI_ATTRIBUTE_NAME, isTouchInterface ? "touch" : "standard");
  }

  private _setFavicon(): void {
    const header = SsrUtil.window?.document.querySelector("head");
    const oldFavicon = SsrUtil.window?.document.querySelector("head link[type='image/x-abstract-icon']");
    let link = SsrUtil.window?.document.createElement("link");
    if (isNotNullNorUndefined(oldFavicon)) {
      link = oldFavicon as any;
    }
    this._renderer.setAttribute(link, "type", "image/x-abstract-icon");
    this._renderer.setAttribute(link, "rel", "shortcut abstract-icon");
    this._renderer.setAttribute(link, "href", `./assets/themes/${this.environment.theme}/favicon.ico`);
    this._renderer.appendChild(header, link);
  }

  private _setImageToolbar(): void {
    this.logo = `assets/themes/${this.environment.theme}/toolbar-header-image.png`;
  }

  logout(): void {
    this._store.dispatch(new SolidifyAppAction.Logout());
  }

  darkThemeChange(enabled: boolean): void {
    this._store.dispatch(new SolidifyAppAction.ChangeDarkMode(enabled));
  }

  themeChange(theme: ExtendEnum<ThemePartialEnum>): void {
    if (theme !== MemoizedUtil.selectSnapshot(this._store, this.environment.appState, state => state.theme)) {
      this._store.dispatch(new SolidifyAppAction.ChangeAppTheme(theme));
    }
  }

  navigateToHome(): void {
    this.navigate(this.environment.routeHomePage);
  }

  private _observeThemeChange(): Observable<ExtendEnum<ThemePartialEnum>> {
    return this.themeObs.pipe(
      distinctUntilChanged(),
      tap((theme: ExtendEnum<ThemePartialEnum>) => {
        this._setTheme();
        this._setFavicon();
        this._setImageToolbar();
      }),
    );
  }

  private _observeDarkModeChange(): Observable<boolean> {
    return this.darkModeObs.pipe(
      distinctUntilChanged(),
      tap((enabled: boolean) => {
        this._setDarkMode(enabled);
      }),
    );
  }

  private _listenLanguageChange(): Observable<ExtendEnum<LanguagePartialEnum>> {
    return this.appLanguageObs.pipe(
      distinctUntilChanged(),
      tap((language: ExtendEnum<LanguagePartialEnum>) => {
        this._findMetaToUpdate();
      }),
    );
  }

  private _listenAppReady(): Observable<boolean> {
    return this.isApplicationInitializedObs.pipe(
      tap((isInit: boolean) => {
        this._findMetaToUpdate();
      }),
    );
  }

  private _setOverrideCssStyleSheet(): void {
    const header = SsrUtil.window?.document.querySelector("head");
    const link = SsrUtil.window?.document.createElement("link");
    this._renderer.setAttribute(link, "type", "text/css");
    this._renderer.setAttribute(link, "rel", "stylesheet");
    this._renderer.setAttribute(link, "href", `./assets/styles/override.css`);
    this._renderer.appendChild(header, link);
  }

  protected _getValueScriptDiffer(script: Element): string | undefined {
    if (isNullOrUndefined(script) || isNullOrUndefined(script.attributes)) {
      return undefined;
    }
    const attribute = script.attributes[this._ATTRIBUTE_SRC_DIFFER_KEY];
    return isNullOrUndefined(attribute) ? undefined : attribute.value;
  }

  protected _getScriptWithValueEndWith(listScript: Element[], threeDimensionalLib: string): Element | undefined {
    return listScript.find(script => {
      const value = this._getValueScriptDiffer(script);
      return isNotNullNorUndefined(value) && value.endsWith(threeDimensionalLib);
    });
  }

  scrollHandler($event: Event): void {
    this._scrollService.triggerScrollEventMain($event);
  }

  private _initDarkMode(): void {
    const darkModeInStorage = LocalStorageHelper.getItem(LocalStoragePartialEnum.darkMode);
    let darkMode = this.environment.darkMode;
    if (darkModeInStorage === SOLIDIFY_CONSTANTS.STRING_TRUE) {
      darkMode = true;
    } else if (darkModeInStorage === SOLIDIFY_CONSTANTS.STRING_FALSE) {
      darkMode = false;
    }
    this.darkThemeChange(darkMode);
  }
}
